(function ($) {
    "use strict";
    /* --------------------------------------------> KEY SETTINGS <-------------------------------------------- */
    var keyAppSettings = 'appSettings';

    var appVersionAndroid = '3.0.2';
    var appVersionIos = '3.0.2';

    /* URLs / Paths for cyberport shop */
    var shopUrlHome = 'de.html';
    var shopUrlSearch = 'de/tools/search-results.html?autosuggest=false&q=';
    var shopUrlAppRatingForm  = 'de/service/hotlines-und-kontakt/feedback-zur-app.html';
    var shopUrlAppSettings  = '/de/appsettings.html';
    /* Indicator if App Settings Page is opened */
    var bodyClassAppSettings = 'appSettings';

    /* URLs for platform specific stores */
    var appUrlAndroidStore = 'market://details?id=de.cyberport.android';
    var appUrlIosStore = 'itms-apps://itunes.apple.com/de/app/cyberport-elektronik-technik-deals-shopping/id561663265?mt=8';
    /* --------------------------------------------> KEY SETTINGS <-------------------------------------------- */

    var appSettingsEmpty = {
        selectedShop : '',
        autoLogin: true,
        fingerprintLogin: false,
        fingerprintLoginDecision: '',
        credentials : {
            username: '',
            password: '',
            valid: false
        },
        lastStartup: 0,
        autoLoginTimeout: 30,
        preventAutoLoginForLoggedOutUser: false,
        countForSessionTimeoutMessage: 0,            // Session Timeout Message is shown, when value = 1
        showDialogAskForFingerprintDecision: true
    };

    /**
     * Workflow:
     * initialize
     *      registerCallbacks
     *      bindEvents
     *          onDeviceReady
     *              init
     *                  handleFirstAppStartup
     *                  handleSubsequentAppStartups
     *                  handleAutologin
     *                  handleFingerprintLogin
     *                  ...
     **/
    var app = {
        initialize: function () {
            var consoleLogger = function (msg) {
                //window.console.log(msg);
            };
            var alertLogger = function (msg) {
                window.alert(msg);
            };
            app.log = app.getLogger(consoleLogger, 'DEBUG');
            app.bindEvents();
            app.registerCallbacks();

            // prevent wkWebView from enabling pinch-to-zoom:
            $('head').append('<meta name="viewport" content="user-scalable=no, initial-scale=1.0, maximum-scale=1.0, minimum-scale=1.0, width=device-width" />');
        },

        /**
         * Used to bind events ondeviceReady and login/logout
         */
        bindEvents: function () {
            $(window).on('cp.logoutSuccess', app.onLogoutSuccess);
            document.addEventListener('deviceready', app.onDeviceReady, false);
            $(window).on('cp.loginFailed', app.onLoginFailed);
            $(window).on('cp.loginSuccessful', app.onLoginSuccessful);
            $('.main-header').find('.logout').click(app.onLogoutClicked);
        },

        /**
         * When user is not on app landing page, function sets callback on login form at user flyout to persist credentials in app settings
         */
        registerCallbacks: function () {
            var currentUrl = window.location.href;
            if (!isAppLandingPage(currentUrl)) {          // -1 never occurs, 1 exists
                window.login.setAppCallbackFunction(persistUserCredentialsFromUi);
            }
        },

        /**
         *
         */
        getLogger: function (logger, level) {
            var logError, logWarning, logInfo, logDebug;
            if (level === 'ERROR') {
                logError = true;
                logWarning = false;
                logInfo = false;
                logDebug = false;
            } else if (level === 'WARNING') {
                logError = true;
                logWarning = true;
                logInfo = false;
                logDebug = false;
            } else if (level === 'INFO') {
                logError = true;
                logWarning = true;
                logInfo = true;
                logDebug = false;
            } else if (level === 'DEBUG') {
                logError = true;
                logWarning = true;
                logInfo = true;
                logDebug = true;
            } else {
                logError = true;
                logWarning = false;
                logInfo = false;
                logDebug = false;
            }

            var loggerInstance = {
                error: function (msg) {
                    if (logError) {
                        logger('ERROR: ' + msg);
                    }
                },
                warning: function (msg) {
                    if (logWarning) {
                        logger('WARNING: ' + msg);
                    }
                },
                info: function (msg) {
                    if (logInfo) {
                        logger('INFO: ' + msg);
                    }
                },
                debug: function (msg) {
                    if (logDebug) {
                        logger('DEBUG: ' + msg);
                    }
                }
            };
            return loggerInstance;
        },

        dialogs: {
            /**
             * Shows an action sheet with an option to cancel the dialog.
             * @param opts
             * {
             *      title : string,     // optional
             *      subtitle : string,  // optional
             *      buttons : []
             * }
             * @param cb Callback function. Callback gets the index of clicked button as parameter.
             */
            showActionSheetWithCancel: function (opts, cb) {
                opts.cancel = true;
                app.dialogs.showActionSheet(opts, cb);
            },

            /**
             * Shows an action sheet.
             * @param opts
             * {
             *      title : string,     // optional
             *      subtitle : string,  // optional
             *      buttons : []
             * }
             * @param cb Callback function. Callback gets the index of clicked button as parameter.
             */
            showActionSheet: function (opts, cb) {
                var options = {
                    androidTheme: window.plugins.actionsheet.ANDROID_THEMES.THEME_DEVICE_DEFAULT_LIGHT, // material design
                    buttonLabels: opts.buttons,
                    //addDestructiveButtonWithLabel : 'Delete it',
                    destructiveButtonLast: true // you can choose where the destructive button is shown
                };
                if (opts.cancel) {
                    options.addCancelButtonWithLabel = 'Abbrechen';
                    options.androidEnableCancelButton = true;
                }
                if (opts.title) {
                    options.title = opts.title;
                }
                if (opts.subtitle) {
                    options.subtitle = opts.subtitle; // supported on iOS only
                }
                window.plugins.actionsheet.show(options, cb);
            },

            /**
             * Shows a confirmation dialog with optional title, optional message and buttons. Callback gets the index of clicked button.
             * @param opts
             * {
             *      title : string,     // optional
             *      message : string,   // optional
             *      buttons : [],
             * }
             * @param cb Callback function
             */
            showConfirmationDialog: function (opts, cb) {
                var message = opts.message ? opts.message : '';
                var title = opts.title ? opts.title : '';
                navigator.notification.confirm(message, cb, title, opts.buttons);
            },

            /**
             * Shows a native notification alert
             * @param title Window title
             * @param message Message to display
             * @param buttonName Name of the button to confirm message
             * @param cb Callback function that is executed when dialog is confirmed
             */
            showNotificationAlert: function(title, message, buttonName, cb) {
                navigator.notification.alert(message, cb, title, buttonName);
            }
        },

        /**
         *  add(): Register a handler with a unique name. If ne name is already registered, the callback function will not be added (no duplicates possible)
         *
         *  remove(): Remove a handler identified by its name
         *
         *  get(); get all handler functions
         */
        loginFailedHandler: {
            handlers: {},
            add: function(name, handler) {
                if (typeof  handler === 'function') {
                    this.handlers[name] = handler;
                }
            },
            remove: function (name) {
                delete this.handlers[name];
            },
            get: function() {
                return this.handlers;
            }
        },

        /**
         * Invalidates credentials in app settings when login failed
         */
        onLoginFailed: function () {
            app.log.debug('cp.loginFailed fired!');
            /* Invalidate credentials */
            getAppSettings(function (err, res) {
                if (!err) {
                    var settings = res;
                    settings.credentials.valid = false;
                    app.log.debug('fired: onLoginFailed; credentials: ' + settings.credentials.valid);
                    updateAppSettings(settings, function (err, res) {
                        if(err) {
                            app.log.error('onLoginFailed() unable to persist credentials. err -> ' + JSON.stringify(err));
                        } else {
                            app.log.debug('onLoginFailed(); credentials saved: ' + res.credentials.valid);
                        }
                    });
                } else {
                    app.log.error('onLoginFailed() unable to read app settings. err -> ' + JSON.stringify(err));
                }
            });

            /* Handle Analytics: Fingerprintlogin failed */
            if(!isUserLoggedIn() && isFlagFingerprintLoginAttemptPerformed()){
                window.ADB.trackAction("login.Fingerprint", {"status":"Fehlgeschlagen - Falsche Zugangsdaten"});
                setFlagFingerprintloginAttemptPerformed(false);
            }
            /* Handle error handlers and messages */
            var errorHandlers = app.loginFailedHandler.get();

            if (Object.keys(errorHandlers).length === 0) {
                app.dialogs.showNotificationAlert('Login fehlgeschlagen', 'Eine Anmeldung ist mit diesen Benutzerdaten nicht möglich.', 'OK');
            } else {
                Object.keys(errorHandlers).forEach(function (handlerName) {
                    app.log.debug("errorhandler: "+handlerName);
                    if (typeof  errorHandlers[handlerName] === 'function') {
                        errorHandlers[handlerName]();
                        app.loginFailedHandler.remove(handlerName);
                    }
                });
            }
        },
        /**
         *
         */
        onLogoutSuccess: function() {
            app.log.info('cp.logoutSuccess fired!'); //
        },

        /**
         * Executed after successful login
         */
        onLoginSuccessful: function() {
            /* Handling for handleAskForFingerprintDecisionDialog on checkout page, because there is no page reload */
            if(isPageCheckoutAccount()){
                getAppSettings(function (err, settings) {
                    if(!err) {
                        handleAskForFingerprintDecisionDialog(settings);
                    } else {
                        app.log.error('onLoginSuccessful(): Unable to read appSettings from native storage! err -> ' + JSON.stringify(err));
                    }
                });
            }
        },
        /**
         * Disables autologin and ask for fingerprintdecision in app settings when user clicked logout.
         */
        onLogoutClicked: function() {
            app.autoLogin.disableForLoggedOutUser();
            app.log.info('onLogoutClicked()');
            getAppSettings(function (err, settings) {
                if(!err) {
                    settings.preventAutoLoginForLoggedOutUser = true;
                    settings.showDialogAskForFingerprintDecision = false;
                    updateAppSettings(settings, function (err, res) {
                        if(err) {
                            app.log.error('onLogoutClicked(): Unable to persist app settings. err -> ' + JSON.stringify(err));
                        } else {
                            app.log.info('onLogoutClicked(): Persisted logout');
                        }
                    });
                } else {
                    app.log.error('onLogoutClicked(): Unable to read appSettings from native storage! err -> ' + JSON.stringify(err));
                }
            });
        },

        onDeviceReady: function () {
            app.log.info('Device is ready.');
            appVersionAndroid = window.BuildInfo.version;
            appVersionIos = window.BuildInfo.version;
            showBackButtonIos();
            init();
        },

        autoLogin: {
            disableForLoggedOutUser: function () {
                window.localStorage.setItem('appEnableAutoLogin', 'false');
                // Prevents Automatic Fingerprint-Verification after Logout for the Session
                app.fingerprintLogin.setFingerprintLoginAttemptOnStartupPerformed(true);
            },
            enableForLoggedOutUser: function () {
                window.localStorage.setItem('appEnableAutoLogin', 'true');
            },
            isEnabledForLoggedOutUser: function () {
                var show = window.localStorage.getItem('appEnableAutoLogin');
                if (show === 'true' || show === null) {
                    return true;
                } else {
                    return false;
                }
            },

            setExpired: function(bool) {
                window.sessionStorage.setItem('autoLoginExpired', '' + bool);
                app.log.debug("setExpired(): " +bool);
            },

            isExpired: function() {
                var expired = window.sessionStorage.getItem('autoLoginExpired');
                return (expired === null || expired === 'true') ?  true : false;
            },

            showExpiredMessage: function(settings){
                if(settings.countForSessionTimeoutMessage===1 && settings.credentials.username && settings.credentials.password){
                    return true;
                } else{
                    return false;
                }
            },

            setShowFingerprintWhenExpired: function (bool) {
                window.sessionStorage.setItem('showFingerprintWhenExpired', '' + bool);
            },

            isShowFingerprintWhenExpired: function () {
                var show = window.sessionStorage.getItem('showFingerprintWhenExpired');
                return (show === null || show === 'true') ?  true : false;
            },
        },

        fingerprintLogin : {
            isAskForFingerprintDecision: function() {
                var ask = window.sessionStorage.getItem('askForFingerprint');
                return (ask === null || ask === 'true') ?  true : false;
            },
            setAskForFingerprintDecision: function(bool) {
                window.sessionStorage.setItem('askForFingerprint', '' + bool);
                getAppSettings(function (err, res) {
                    if (!err) {
                        var settings = res;
                        settings.showDialogAskForFingerprintDecision = bool;
                        updateAppSettings(settings, function (err, upd) {
                            if(err) {
                                app.log.error('setAskForFingerprintDecision unable to persist credentials. err -> ' + JSON.stringify(err));
                            }
                        });
                    } else {
                        app.log.error('setAskForFingerprintDecision unable to read app settings. err -> ' + JSON.stringify(err));
                    }
                });
            },
            askForFingerprint: function(cb){
                var opts;
                if (isAndroid()) {
                    opts = {
                        title: 'Fingerabdruck Login',
                        subtitle: 'Schnell und bequem per Fingerabdruck einloggen? Ihre Anmeldedaten werden sicher verschlüsselt im Schlüsselbund Ihres Gerätes gespeichert.',
                        buttons: ['Ja', 'Später', 'Nein']
                    };
                } else if (isIos()) {
                    opts = {
                        title: 'Touch ID aktivieren',
                        subtitle: 'Schnell und bequem per Fingerabdruck einloggen? Ihre Anmeldedaten werden sicher verschlüsselt im Schlüsselbund Ihres Gerätes gespeichert.',
                        buttons: ['Ja', 'Später', 'Nein']
                    };
                } else {
                    app.log.error('Plattform für Fingerprintlogin nicht erkannt.');
                }
                app.dialogs.showActionSheet(opts, cb);
            },
            setFingerprintLoginAttemptOnStartupPerformed: function(bool) {
                window.sessionStorage.setItem('fingerprintLoginAttemptOnStartupPerformed', '' + bool);
            },
            isFingerprintLoginAttemptOnStartupPerformed: function() {
                var performed = window.sessionStorage.getItem('fingerprintLoginAttemptOnStartupPerformed');
                return (performed === 'true') ?  true : false;
            },

        },
        /**
         * Function used to display and bind app rating button on app settings.
         */
        triggerAppRatingOpenMenuEvent: function(){
            var opts = {
                title: 'Bewerten Sie unsere App',
                subtitle: 'Ihr Feedback ist uns wichtig! Gemeinsam mit Ihnen wollen wir die App stetig verbessern. Gefällt Ihnen die Cyberport App?',
                buttons: ['Geben Sie uns 5 Sterne','Melden Sie uns Ihr Feedback','Abbrechen']
            };
            var appRatingActionCallback = function (buttonIndex) {
                // ButtonIndex
                // 1: 5 Sterne,
                // 2: Feedback Formular,
                // 3: Später
                if (buttonIndex === 1) {
                    if (isAndroid()) {
                        window.ADB.trackAction("setting.Apprating", {"rating":"5 Sterne Android"});
                        app.log.info('Debug:AppRatingEvent:Android');
                        window.location.href = appUrlAndroidStore;
                    } else if (isIos()) {
                        window.ADB.trackAction("setting.Apprating", {"rating":"5 Sterne iOS"});
                        app.log.info('Debug:AppRatingEvent:iOS');
                        window.location.href = appUrlIosStore;
                    } else {
                        app.log.error('Could not detect device platform / OS.');
                    }
                } else if (buttonIndex === 2) {
                    window.ADB.trackAction("setting.Apprating", {"rating":"Feedback geben"});
                    var urlFeedbackForm = getBaseUrl() + shopUrlAppRatingForm;
                    app.log.info("Feedback Form URL: " + getBaseUrl() + shopUrlAppRatingForm );
                    window.location.href = urlFeedbackForm;
                } else if (buttonIndex === 3) {
                    window.ADB.trackAction("setting.Apprating", {"rating":"Abbruch"});
                    window.plugins.actionsheet.hide();
                }
            };
            app.dialogs.showActionSheet(opts, appRatingActionCallback);
        },

        /**
         * Function executed when user clicked on barcodescanner in search.
         */
        triggerBarcodeScannerEvent: function() {
            window.ADB.trackTimedActionStart("barcodeScanner");
            var startBarcodeScanTime = new Date();
            window.cordova.plugins.barcodeScanner.scan(
                function (result) {
                    var timeDiffBarcodeScanTime = ((new Date()  - startBarcodeScanTime )/1000);
                    window.ADB.trackAction("barcodeScanner", {"duration":timeDiffBarcodeScanTime});
                    window.ADB.trackTimedActionEnd("barcodeScanner");
                    app.log.debug("timeDiffBarcodeScanTime "+timeDiffBarcodeScanTime);
                    if (!result.cancelled) {
                        var url = getBaseUrl() + shopUrlSearch + result.text;
                        window.ADB.trackAction("barcodeScanner", {"landingPage":url});
                        window.location.href = url;
                    } else {
                        window.ADB.trackAction("barcodeScanner", {"landingPage":"none: abbruch"});
                    }
                },
                function (error) {
                    if(error==="Illegal access" || isIos()){
                        if (window.cordova && window.cordova.plugins.settings) {
                            var barcodeScannerPermissionCallback = function (buttonIndex) {
                                if (buttonIndex === 2) {
                                    if(isAndroid()){
                                        window.cordova.plugins.settings.open("application_details", function() {
                                            },
                                            function () {
                                                app.log.debug('failed to open settings');
                                            }
                                        );
                                    } else {
                                        window.cordova.plugins.settings.open("settings", function() {
                                            },
                                            function () {
                                                app.log.debug('failed to open settings');
                                            }
                                        );
                                    }
                                }
                            }
                            var timeDiffBarcodeScanTime = Math.round(((new Date()  - startBarcodeScanTime )/1000) % 60);
                            window.ADB.trackAction("barcodeScanner", {"landingPage":"none: keine Kameraberechtigung"});
                            window.ADB.trackAction("barcodeScanner", {"duration":timeDiffBarcodeScanTime})
                            window.ADB.trackTimedActionEnd("barcodeScanner");

                            navigator.notification.confirm(
                                'Cyberport hat keinen Zugriff auf Ihre Kamera. Um Zugriff zu erlauben, Geräte-Einstellungen öffnen und unter Cyberport die Verwendung der Kamera aktivieren.',  // message
                                barcodeScannerPermissionCallback,            // callback to invoke
                                'Scan nicht möglich',                       // title
                                ['Abbrechen','Einstellungen']             // buttonLabels
                            );
                        } else {
                            var timeDiffBarcodeScanTime = Math.round(((new Date()  - startBarcodeScanTime )/1000) % 60);
                            window.ADB.trackTimedActionEnd("barcodeScanner");
                            window.ADB.trackAction("barcodeScanner", {"landingPage":"none: keine Kameraberechtigung"});
                            window.ADB.trackAction("barcodeScanner", {"duration":timeDiffBarcodeScanTime});
                            app.dialogs.showNotificationAlert('Scan nicht möglich', 'Cyberport hat keinen Zugriff auf Ihre Kamera. Um Zugriff zu erlauben, Geräte-Einstellungen öffnen und unter Cyberport die Verwendung der Kamera aktivieren', 'OK');
                        }
                    } else {
                        var timeDiffBarcodeScanTime = Math.round(((new Date()  - startBarcodeScanTime )/1000) % 60);
                        window.ADB.trackTimedActionEnd("barcodeScanner");
                        window.ADB.trackAction("barcodeScanner", {"landingPage":"none: keine Kameraberechtigung"});
                        window.ADB.trackAction("barcodeScanner", {"duration":timeDiffBarcodeScanTime});
                        app.dialogs.showNotificationAlert('Scan nicht möglich', 'Leider hat der Scan nicht funktioniert.', 'OK');
                    }
                    app.log.error('Barcodescan not successful. Error -> ' + JSON.stringify(error));
                },
                {
                    preferFrontCamera : false, // iOS and Android
                    showFlipCameraButton : false, // iOS and Android
                    showTorchButton : true, // iOS and Android
                    torchOn: false, // Android, launch with the torch switched on (if available)
                    prompt : 'Platzieren Sie den Barcode in den sichtbaren Bereich', // Android
                    resultDisplayDuration: 0, // Android, display scanned text for X ms. 0 suppresses it entirely, default 1500
                    formats : 'EAN_8,EAN_13', // default: all but PDF_417 and RSS_EXPANDED
                    orientation : 'portrait', // Android only (portrait|landscape), default unset so it rotates with the device
                    disableAnimations : false, // iOS
                    disableSuccessBeep: true // iOS
                }
            );
        }
    };
    /**
     * Persists user credentials from ui / flyout.
     */
    function persistUserCredentialsFromUi() {
        var credentials = window.login.getLoginMailPassword();
        if (credentials) {
            if (credentials.email && credentials.password) {
                getAppSettings(function (err, res) {
                    if (!err) {
                        var settings = res;
                        settings.credentials.username = credentials.email;
                        settings.credentials.password = credentials.password;
                        settings.credentials.valid = true;
                        settings.preventAutoLoginForLoggedOutUser = false;
                        settings.showDialogAskForFingerprintDecision = true;
                        updateAppSettings(settings, function (err, res) {
                            if(err) {
                                app.log.error('persistUserCredentialsFromUi unable to persist credentials. err -> ' + JSON.stringify(err));
                            }
                        });
                    } else {
                        app.log.error('persistUserCredentialsFromUi unable to read app settings. err -> ' + JSON.stringify(err));
                    }
                });
            }
        }
    }

    /**
     * Returns the status of the fingerprint sensor. Result:
     * {
     *  isAvailable: boolean
     *  isHardwareDetected:	boolean
     *  hasEnrolledFingerprints: boolean
     *  type: For Apple devices: 'face' on iPhone X, 'touch' on other apple devices; Not set for Android
     * }
     * @param cb
     */
    function isFingerprintScannerAvailable(cb){
        if (isAndroid()) {
            window.FingerprintAuth.isAvailable(function (result) {
                //app.log.debug('Fingerprint available? res -> ' + JSON.stringify(result));
                cb(null, result);
            }, function(error) {
                //app.log.error('error checking Fingerprint -> ' + JSON.stringify(error));
                cb(error, null);
            });
        } else if (isIos()) {
            window.plugins.touchid.isAvailable(function(type) {
                //app.log.info('Fingerprint available!');
                var result = {};
                result.isAvailable = true;
                result.isHardwareDetected = true;
                result.hasEnrolledFingerprints = true;
                if(type){
                    result.type = type;
                }
                cb(null, result);
            }, function(res) {
                var result = {};
                // Status Codes iOS: Keine Finger registriert: -7, kein TouchID verbaut: -6
                if (res.code === -7) {
                    result.isAvailable = false;
                    result.isHardwareDetected = true;
                    result.hasEnrolledFingerprints = false;
                    cb(null, result);
                } else if (res.code === -6){
                    result.isAvailable = false;
                    result.isHardwareDetected = false;
                    result.hasEnrolledFingerprints = false;
                    cb(null, result);
                } else {
                    // TODO maybe use a common error Object for iOS and Android
                    cb(res, null);
                }
            });
        } else {
            cb(new Error('Unsupported platform. Only iOS and Android are supported.'), null);
        }
    }

    /**
     * Function verifies fingerprint
     * @param {}
     */
    function verifyFingerprint(cb) {
        var errorObj = {};
        errorObj.isDialogCanceled = false;
        errorObj.code = 0;
        errorObj.message = '';

        if (isAndroid()) {
            var encryptConfig = {
                clientId: 'de.cyberport.mobile.aem.app',
                //username: 'currentUser',
                //password: 'currentUserPassword',
                locale: 'de',
                disableBackup: true, // TODO Disable PIN entry screen since it is not working currently. See https://github.com/mjwheatley/cordova-plugin-android-fingerprint-auth/issues/85
                dialogTitle: 'Einloggen mit Fingerabdruck',
                dialogMessage: 'Scannen Sie Ihren Fingerabdruck, um auf Kundenkonto, Warenkorb und die Merkliste zuzugreifen.',
            };

            var successCallbackAndroid = function (result) {
                app.log.info('successCallbackAndroid(): ' + JSON.stringify(result));
                if (result.withFingerprint) {
                    app.log.info('Successfully encrypted credentials.');
                    app.log.info('Encrypted credentials: ' + result.token);
                    cb(null, true);
                } else if (result.withBackup) {
                    app.log.info('Authenticated with backup password');
                    cb(null, true);
                }
            };

            var errorCallbackAndroid = function (error) {
                if (error === window.FingerprintAuth.ERRORS.FINGERPRINT_CANCELLED) {
                    app.log.info('FingerprintAuth Dialog Cancelled!');
                    window.ADB.trackAction("login.Fingerprint", {"status":"Fehlgeschlagen - Abbruch durch User"});
                    errorObj.isDialogCanceled = true;
                    errorObj.code = -1;
                    errorObj.message = error;
                    cb(errorObj, false);
                } else {
                    app.log.error('FingerprintAuth Error1: ' + error);
                    window.ADB.trackAction("login.Fingerprint", {"status":"Fehlgeschlagen"});
                    errorObj.isDialogCanceled = false;
                    errorObj.code = -1;
                    errorObj.message = error;
                    cb(errorObj, false);
                }
            };

            window.FingerprintAuth.encrypt(encryptConfig, successCallbackAndroid, errorCallbackAndroid);

        } else if (isIos()) {
            var successCallbackIos = function () {
                cb(null, true);
            };

            var errorCallbackIos = function (res) {
                // res -> {
                //          localizedDescription: string,
                //          code: int // bei -128 wurde der dialog abgebrochen
                //         }
                if (res.code === -128) {
                    app.log.info('FingerprintAuth Dialog Cancelled!');
                    window.ADB.trackAction("login.Fingerprint", {"status":"Fehlgeschlagen - Abbruch durch User"});
                    errorObj.isDialogCanceled = true;
                    errorObj.code = res.code;
                    errorObj.message = res.localizedDescription;
                    cb(false, errorObj);
                } else {
                    window.ADB.trackAction("login.Fingerprint", {"status":"Fehlgeschlagen"});
                    errorObj.isDialogCanceled = false;
                    errorObj.code = res.code;
                    errorObj.message = res.localizedDescription;
                    app.log.error('FingerprintAuth error: ' + JSON.stringify(errorObj));
                    cb(false, errorObj);
                }
            };

            // TODO maybe add the check mentioned here: https://github.com/EddyVerbruggen/cordova-plugin-touch-id#security
            window.plugins.touchid.verifyFingerprint('Scannen Sie Ihren Fingerabdruck, um auf Kundenkonto, Warenkorb und die Merkliste zuzugreifen.', successCallbackIos, errorCallbackIos);
        } else {
            app.log.error('Fingerprint Scanner is not supported on Your current device platform.');
        }
    }
    /**
     * @returns {String} ios, android or other
     */
    function getMobilePlatform() {
        if (!!navigator.platform && /iPad|iPhone|iPod/.test(navigator.platform)) {
            return 'ios';
        } else if (navigator.userAgent.toLowerCase().indexOf('android') > -1) {
            return 'android';
        } else {
            return 'other';
        }
    }
    /**
     * @returns {boolean}
     */
    function isAndroid(){
        return 'android' === getMobilePlatform();
    }
    /**
     * @returns {boolean}
     */
    function isIos() {
        return 'ios' === getMobilePlatform();
    }
    /**
     * @returns {boolean}
     */
    function isIpad(){
        return navigator.userAgent.match(/iPad/i) != null;
    }

    /**
     * Detects via viewport width whether the device is a tablet or not.
     * @returns {boolean}
     */
    function isTablet(){
        var viewportWidth = Math.max(document.documentElement.clientWidth, window.innerWidth || 0);
        if(viewportWidth>767){
            return true;
        } else {
            return false;
        }
    }
    /**
     * Detects via viewport width/height whether the device is a iphone x.
     * @returns {boolean}
     */
    function isIphoneX(){
        var ratio = window.devicePixelRatio || 1;
        var screen = {
            width : window.screen.width * ratio,
            height : window.screen.height * ratio
        };
        if (isIos() && screen.width == 1125 && screen.height === 2436) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * Init function handling
     */
    function init() {
        getAppSettings(function (err, settings) {
            if(!err) {
                initUniversalLinks(settings);
            } else {
                app.log.error('onDeviceReady(): Unable to read appSettings from native storage! err -> ' + JSON.stringify(err));
            }
        });
        var currentUrl = window.location.href;
        if (isAppLandingPage(currentUrl)) {          // -1 never occurs, 1 exists
            checkAppSettings(function (err, res) {
                if (!err) {
                    if (res === true) {
                        handleSubsequentAppStartups();
                    } else {
                        handleFirstAppStartup();
                    }
                } else {
                    app.log.error('init()/checkAppSettings(): Unable to read appSettings from device on app landing page! err -> ' + JSON.stringify(err));
                }
            });
        } else {
            $('a').css('-webkit-touch-callout', 'none');
            if(isIpad() || isTablet()){
                $( '#tooltipBoxAccount .showInApp' ).each(function () {
                    this.style.setProperty( 'display', 'initial', 'important' );
                });
            }
            navigator.splashscreen.hide();

            // this is a workaround if user changes the selected shop in app settings an then uses the "back" button of browser/app
            // selected_shop cookie and URL do not match -> redirect to homepage of shop
            if (isUrlWithInvalidLocale(currentUrl)) {
                window.location.replace(getBaseUrl() + shopUrlHome);
            }

            // Handle Analytics: Log successfull fingerprintlogin
            if(isUserLoggedIn() && isFlagFingerprintLoginAttemptPerformed()){
                window.ADB.trackAction("login.Fingerprint", {"status":"Erfolgreich"});
                setFlagFingerprintloginAttemptPerformed(false);
            }

            getAppSettings(function (err, settings) {
                if(!err) {
                    // Is called, when user clicked on logout.
                    if(!app.autoLogin.isEnabledForLoggedOutUser()){
                        settings.preventAutoLoginForLoggedOutUser = true;
                        updateAppSettings(settings, function (err, res) {
                            if(err) {
                                app.log.error('init(): Unable to persist app settings. err -> ' + JSON.stringify(err));
                            }
                        });
                    }
                    handleAutoLogin(settings);
                    handleFingerprintLogin(settings, currentUrl);
                } else {
                    app.log.error('init(): Unable to read appSettings from native storage! err -> ' + JSON.stringify(err));
                }
            });
            initAppSettingsPage();
            initBarcodeScanner();
            initFeedbackFormPage();
        }
    }

    /**
     * Function handles auto login.
     * @param {} settings - local app settings.
     */
    function handleAutoLogin(settings) {
        if (!isUserLoggedIn()) {
            if(app.autoLogin.isExpired()){
                app.log.debug("handleAutoLogin: Autologin is expired");
                settings.preventAutoLoginForLoggedOutUser = true;
                updateAppSettings(settings, function (err, updatedSettings) {
                    if (err) {
                        app.log.error('handleAutoLogin isExpired: Unable to persist updated appSettings after failed autoLogin. err -> ' + JSON.stringify(err));
                    }
                });
            }
            if (!app.autoLogin.isExpired() && settings.autoLogin === true && !settings.preventAutoLoginForLoggedOutUser && settings.credentials.username && settings.credentials.password && settings.credentials.valid) {
                var loginContainer = getLoginComponentHeaderBarNoReload();

                // add handler to remove loading animation of login component and invalidate credentials on error.
                app.loginFailedHandler.add('errorAutoLogin', function () {
                    app.dialogs.showNotificationAlert('Login fehlgeschlagen', 'Ihr gespeichertes Passwort ist nicht mehr gültig. Bitte melden Sie sich erneut an.', 'OK');
                    window.animationHelper.removeLoadingAnimation(loginContainer.find('.component.login'));
                    settings.credentials.valid = false;
                    updateAppSettings(settings, function (err, updatedSettings) {
                        if (err) {
                            app.log.error('init: Unable to persist updated appSettings after failed autoLogin. err -> ' + JSON.stringify(err));
                        }
                    });
                });
                loginWithInvisibleForm(loginContainer, settings.credentials.username, settings.credentials.password);
            } else {
                var loginHeader = getLoginComponentHeaderBar();
                var loginCheckout = getLoginComponentCheckoutProcess();
                var loginWishlist = getLoginComponentWishList();
                var username = settings.credentials.username;

                if (loginHeader.length !== 0) {
                    loginHeader.find('form.loginForm').find('input[name="j_username"]').val(username);
                }
                if (loginCheckout.length !== 0) {
                    loginCheckout.find('form.loginForm').find('input[name="j_username"]').val(username);
                }
                if (loginWishlist.length !== 0) {
                    loginWishlist.find('form.loginForm').find('input[name="j_username"]').val(username);
                }
            }
        } else {
            if(settings.preventAutoLoginForLoggedOutUser){
                settings.preventAutoLoginForLoggedOutUser = false;
                app.autoLogin.enableForLoggedOutUser();
                app.log.debug('fired: handleAutoLogin isUserLoggedIn: preventAutoLoginForLoggedOutUser false');
                updateAppSettings(settings, function (err, updatedSettings) {
                    if (err) {
                        app.log.error('init: Unable to persist updated appSettings. err -> ' + JSON.stringify(err));
                    }
                });
            }
        }
    }
    /**
     * Function handles fingerprint login.
     */
    function handleFingerprintLogin(settings, currentUrl) {
        isFingerprintScannerAvailable(function (err, res) {
            if (!err) {
                if (res.isAvailable) {
                    registerEventHandlersForFingerprintLogin(settings, currentUrl);
                    if (isUserLoggedIn()) {
                        handleAskForFingerprintDecisionDialog(settings);
                    } else { // user is not logged in
                        app.log.debug('Fingerprint, User not logged in..');
                        if (settings.fingerprintLogin === true && settings.credentials.username && settings.credentials.password && settings.credentials.valid) {
                            if (!isUserLoggedIn() && (!settings.autoLogin || settings.preventAutoLoginForLoggedOutUser) && ((app.autoLogin.isExpired() && app.autoLogin.isShowFingerprintWhenExpired()) || (!app.fingerprintLogin.isFingerprintLoginAttemptOnStartupPerformed()))) {
                                app.log.debug('Auto Login expired. Showing Fingerprint dialog.');
                                verifyFingerprint(function(err, fingerprintIsValid) {
                                    app.log.debug('attempt to logging in after touchID verification');
                                    app.fingerprintLogin.setFingerprintLoginAttemptOnStartupPerformed(true);
                                    if (!err) {
                                        if (fingerprintIsValid === true) {
                                            var loginComponent = getLoginComponentHeaderBar();
                                            // add handler to remove loading animation of login component
                                            app.loginFailedHandler.add('errorDefaultLogin', function(){
                                                app.dialogs.showNotificationAlert('Login fehlgeschlagen', 'Ihr gespeichertes Passwort ist nicht mehr gültig. Bitte melden Sie sich erneut an.', 'OK');
                                                window.animationHelper.removeLoadingAnimation(loginComponent.find('.component.login'));
                                            });
                                            loginWithInvisibleForm(loginComponent, settings.credentials.username, settings.credentials.password, "fingerprint");
                                        } else {
                                            window.ADB.trackAction("login.Fingerprint", {"status":"Fehlgeschlagen - Fingerprintverifizierung fehlgeschlagen"});
                                        }
                                    } else {
                                        app.log.error('Error attempting to verify fingerprint. err -> ' + JSON.stringify(err));
                                        app.autoLogin.disableForLoggedOutUser();                                    }
                                    // Only show fingerprint dialog one time
                                    app.autoLogin.setShowFingerprintWhenExpired(false);
                                });
                            }
                        }
                    }
                }
            } else {
                app.log.error('handleFingerprintLogin(): Unable to check state of fingerprint scanner. err ->' + JSON.stringify(err));
            }
        });
    }

    /**
     * Function shows non-native back button in iOS
     */
    function showBackButtonIos() {
        var backButton = $('#appHistoryBack');
        var referrer = document.referrer;
        var currentUrl = window.location.href;
        var historyLength = window.history.length;
        if (referrer) {
            if (!isAppLandingPage(referrer) && backButton.length > 0 && isIos() && !isIpad() && !((historyLength<2) && (referrer == currentUrl))) {
                $(backButton).on('click', function () {
                    window.history.back();
                });
                $(backButton).removeClass('showInApp').addClass('is--active');
            }
        }
    }

    /**
     * Function checks if opened URL is feedback form. If feedback form page is opened, function fills form with device and os information.
     */
    function initFeedbackFormPage() {
        var currentUrl = window.location.href;
        if (currentUrl.indexOf('feedback-zur-app.html') > -1) {          // -1 never occurs, 1 exists
            // Set device_name = Name des Gerätes (bspw. iPhone 6s)
            $("input[name='device_name']").val(device.manufacturer + " " + device.model);

            // Set os_version = Name und Version der Betriebssystems (bspw. iOS 10.3.3)
            $("input[name='os_version']").val(device.platform + " " + device.version);

            // Set app_version = Name der App Version (bspw. iOS-App 0.1.2)
            if (isAndroid()) {
                $("input[name='app_version']").val(appVersionAndroid);
            } else if (isIos()) {
                $("input[name='app_version']").val(appVersionIos);
            } else {
                $("input[name='app_version']").val("App-Platform nicht erkannt! (" + appVersionAndroid + " oder " + appVersionIos + ")");
            }
        }
    }

    /**
     * Function for deep linking: DL opens external URLs from browser in app
     */
    function initUniversalLinks(settings){
        universalLinks.subscribe('cpShopMain', function (eventData) {
            app.log.debug("external URL: "+eventData.url);
            var res = settings;
            var path = eventData.url;
            var pathSplit = path.split("/");
            if (pathSplit[1] == "content") {
                // external link comes from app domain app.cyberport.de/content/webshop30_de/de
                path = path.replace("/content/webshop30_de/de", "");
                path = path.replace("/content/webshop30_at/de", "");
                window.location.replace(getBaseUrlForLocale(res.selectedShop) + "de" + path);
            } else {
                /* External link comes from cyberport.de or cyberport.at; Check if user comes from start page.
                 * TRUE: Redirection to app start page (de.html)
                 * FALSE: Redirection to pathname.
                 */
                if (path == "/") {
                    window.location.replace(getBaseUrlForLocale(res.selectedShop) + shopUrlHome);
                } else {
                    var path = path.replace("/pdp", ".pdp");
                    window.location.replace(getBaseUrlForLocale(res.selectedShop) + "de" + path);
                }
            }
            window.location.href = eventData.url;
        });
    }

    /**
     * Function displays app settings link in burger menu and if app settings page is opened, sets values in page.
     * Function removes loading spinner in app settings page.
     */
    function initAppSettingsPage(){
        $('li .showInApp').each(function () {
            this.style.setProperty( 'display', 'list-item', 'important' );
        });
        if(isAppSettingsPage()){
            // Button for App-Rating
            appSettingsPageUi.initAppRatingPopUpFunction();
            // Sets App Version
            appSettingsPageUi.setUiAppVersion();

            // Start App Settings
            getAppSettings(function (err, res) {
                if (!err) {
                    var settings = res;

                    // Country/Shop
                    appSettingsPageUi.setUiShopSwitchCountry(settings);
                    appSettingsPageUi.handleSwitchShopCountry(settings);

                    // Auto Login
                    appSettingsPageUi.setUiAutologin(settings);
                    appSettingsPageUi.handleSwitchAutologin(settings);
                    // Fingerprint
                    isFingerprintScannerAvailable(function (err, fingerPrintScanner) {
                        if (!err) {
                            appSettingsPageUi.setUiFingerprintLogin(settings, fingerPrintScanner);
                            appSettingsPageUi.handleSwitchFingerprintLogin(settings, fingerPrintScanner);
                        }
                    });

                    // Hide Spinner
                    var appSettingsSpinner = $('#appSettingsSpinner');
                    if(appSettingsSpinner){
                        appSettingsSpinner.hide();
                    }
                } else {
                    app.log.error('unable to read app settings. err -> ' + JSON.stringify(err));
                }
            });
        }
    }

    /**
     * Used to display barcodescanner icon and binding of triggerBarcodeScannerEvent()
     */
    function initBarcodeScanner() {
        var scannerIcon = $('#barcodescanner-button');
        if (scannerIcon.length === 1) {

            // make barcode scanner icon visible
            scannerIcon.removeClass('showInApp');

            // bind barcode scanner click event
            scannerIcon.click(function (event) {
                //remove default click event of button
                event.preventDefault();
                app.triggerBarcodeScannerEvent();
            });
        }
    }

    /**
     * Invisible form used for fingerprintlogin or autologin.
     *
     * @param {} loginComponent Cloned login form
     * @param {string} username
     * @param {string} password
     * @param {string} loginPerformedBy logintype, e.g. fingerprint. Used to set if fingerprintlogin was performed on startup.
     */
    function loginWithInvisibleForm(loginComponent, username, password, loginPerformedBy) {

        if(loginPerformedBy){
            if(loginPerformedBy === "fingerprint"){
                setFlagFingerprintloginAttemptPerformed(true);
            }
        }
        getAppSettings(function (err, res) {
            if (!err) {
                var settings = res;
                settings.preventAutoLoginForLoggedOutUser = false;
                app.autoLogin.enableForLoggedOutUser();
                app.log.debug('fired: loginWithInvisibleForm: preventAutoLoginForLoggedOutUser false');

                updateAppSettings(settings, function (err, res) {
                    if(err) {
                        app.log.error('loginWithInvisibleForm unable to persist credentials. err -> ' + JSON.stringify(err));
                    }
                });
            } else {
                app.log.error('loginWithInvisibleForm unable to read app settings. err -> ' + JSON.stringify(err));
            }

            var loginComponentCloned = loginComponent.clone();
            window.login.init(loginComponentCloned);
            var loginForm = loginComponentCloned.find('form.loginForm');
            loginForm.find('input[name="j_username"]').val(username);
            loginForm.find('input[name="j_password"]').val(password);
            window.animationHelper.addLoadingAnimation(loginComponent.find('.component.login'));
            loginComponentCloned.find('form.loginForm').submit();
            handleAskForFingerprintDecisionDialog(settings);
        });

    }

    /**
     * @returns {jQuery|HTMLElement|*}
     */
    function getLoginComponentHeaderBar() {
        return $('#interaction-bar-search div.loginContainer'); // Login Komponente Startseite
    }
    /**
     * @returns {jQuery|HTMLElement|*}
     */
    function getLoginComponentHeaderBarNoReload() {
        return $('#interaction-bar-search div.component.login'); // Login Komponente Startseite
    }
    /**
     * @returns {jQuery|HTMLElement|*}
     */
    function getLoginComponentCheckoutProcess() {
        return $('div.loginRegistration .loginContainer');  // Login Komponente Checkout Prozess
    }
    /**
     * @returns {jQuery|HTMLElement|*}
     */
    function getLoginComponentWishList() {
        return $('#wishlistSaveOverlay div.loginContainer'); // Login Komponente Merkzettel Speichern
    }

    /**
     * Used for registering fingerprintlogin event handlers.
     *
     * @param {} settings Local app settings
     * @param {string} currentUrl
     */
    function registerEventHandlersForFingerprintLogin(settings, currentUrl) {
        if(app.loginFailedHandler.get()===0){
            // Add error message dialog. Will be displayed if for some reason the login with fingerprint fails.
            app.loginFailedHandler.add('errorMessageFingerpring', function(){
                app.dialogs.showNotificationAlert('Login fehlgeschlagen', 'Eine Anmeldung ist mit diesen Benutzerdaten nicht möglich', 'OK');
            });
        }
        app.log.debug('registerEventHandlersForFingerprintLogin fired');
        // Login / Useraccount Flyout
        var res = $('.interaction-bar-account a.main-header-button');
        if (res.length === 1) {
            // Handler darf nur dann feuern, wenn das Flyout nicht geöffnet war (wenn es geöffnet ist, hat es zusätzlich die CSS Klasse is--open )
            res.on('click', function() {
                app.log.debug('registerEventHandlersForFingerprintLogin res.on click fired');
                if ($(this).hasClass('is--open') && !isUserLoggedIn()){
                    app.log.debug('show fingerprint...');
                    getAppSettings(function (err, settings) {
                        if (!err) {
                            if(settings.fingerprintLogin && settings.credentials.username && settings.credentials.password && settings.credentials.valid){
                                verifyFingerprint(function(err, fingerprintIsValid) {
                                    app.log.debug('attempt to logging in after touchID verification');
                                    if (!err) {
                                        if (fingerprintIsValid === true) {
                                            var loginComponent = getLoginComponentHeaderBar();

                                            // add handler to remove loading animation of login component
                                            app.loginFailedHandler.add('errorDefaultLogin', function(){
                                                window.animationHelper.removeLoadingAnimation(loginComponent.find('.component.login'));
                                            });

                                            loginWithInvisibleForm(loginComponent, settings.credentials.username, settings.credentials.password, "fingerprint");
                                        }
                                    } else {
                                        app.log.error('Error trying to verify fingerprint. err -> ' + JSON.stringify(err));
                                    }
                                });
                            }
                        } else {
                            app.log.error('handleSubsequentAppStartups: Unable to read appSettings. error -> ' + JSON.stringify(err));
                        }
                    });
                }
            });
        }
        // Merkzettel speichern (nicht eingeloggt, daher soll TouchID beim Aufrufen der "Speichern" Funktion angezeigt werden.
        if (currentUrl.indexOf('wishlist.html') > -1) {
            var saveListLinkElem = $('.dropdownBox .save .saveList');
            if(saveListLinkElem.length === 1) {
                saveListLinkElem.on('click', function () {
                    if (!isUserLoggedIn() && settings.fingerprintLogin && settings.credentials.username && settings.credentials.password && settings.credentials.valid) {
                        app.log.debug('Show fingerprint!');
                        verifyFingerprint(function(err, fingerprintIsValid) {
                            app.log.debug('attempt to logging in after touchID verification');
                            if (!err) {
                                if (fingerprintIsValid === true) {
                                    var loginComponent = getLoginComponentWishList();

                                    app.loginFailedHandler.add('errorWishlistLogin', function(){
                                        window.animationHelper.removeLoadingAnimation(loginComponent.find('.component.login'));
                                    });

                                    loginWithInvisibleForm(loginComponent, settings.credentials.username, settings.credentials.password, "fingerprint");
                                }
                            } else {
                                app.log.error('Error trying to verify fingerprint. err -> ' + JSON.stringify(err));
                            }
                        });
                    }
                });
            }
            // Warenkorb und Checkout Prozess
        } else if (currentUrl.indexOf('checkout.html') > -1 || currentUrl.indexOf('checkout.cart.html') > -1 ) {
            var elems = document.getElementsByClassName('component checkout');

            if (elems.length === 1) {
                // Observes attribute 'data-current-view' for changes and fires fingerprint dialog if necessary
                var observer = new MutationObserver(function(mutations) {
                    mutations.forEach(function(mutation) {
                        if (mutation.type === 'attributes' && mutation.attributeName === 'data-current-view' && mutation.target.getAttribute('data-current-view') === 'account') {
                            if (!isUserLoggedIn() && settings.fingerprintLogin && settings.credentials.username && settings.credentials.password && settings.credentials.valid) {
                                app.log.debug('show fingerprint!');

                                verifyFingerprint(function(err, fingerprintIsValid) {
                                    app.log.debug('attempt to logging in after touchID');
                                    if (!err) {
                                        if (fingerprintIsValid === true) {
                                            var loginComponent = getLoginComponentCheckoutProcess();

                                            app.loginFailedHandler.add('errorCheckoutProcessLogin', function(){
                                                window.animationHelper.removeLoadingAnimation(loginComponent.find('.component.login'));
                                            });

                                            loginWithInvisibleForm(loginComponent, settings.credentials.username, settings.credentials.password, "fingerprint");
                                        }
                                    } else {
                                        app.log.error('Error trying to verify fingerprint. err -> ' + JSON.stringify(err));
                                    }
                                });
                            }
                        }
                    });
                });

                observer.observe(elems[0], {
                    attributes: true //configure observer to listen to attribute changes
                });
            }
        }
    }
    /**
     * Checks if user is logged in
     * @returns {boolean}
     */
    function isUserLoggedIn() {
        if (window.contextHubHandler) {
            return window.contextHubHandler.isLoggedIn();
        } else {
            return false;
        }
    }

    /**
     * Creates empty app settings on first app start.
     */
    function handleFirstAppStartup(){
        var settings;
        createAppSettings(function (err, res) {
            if (!err) {
                settings = res;
                getLocaleFromDevice(function(err, locale) {
                    if(!err) {
                        settings.selectedShop = getSelectedShopForLocale(locale);

                        updateAppSettings(settings, function(err, updatedObj){
                            if(!err) {
                                handleSubsequentAppStartups();
                            } else {
                                app.log.error('handleFirstAppStartup: Unable to update appSettings. Error -> ' + JSON.stringify(err));
                            }
                        });
                    } else {
                        app.log.error('handleFirstAppStartup: Unable to read locale from device. err -> ' + JSON.stringify(err));
                    }
                });
            } else {
                app.log.error('handleFirstAppStartup: Unable to initialize appSettings. err -> ' + JSON.stringify(err));
            }
        });
        app.autoLogin.enableForLoggedOutUser();
    }
    /**
     * Functions tasks:
     * - Checks if autologin session timed out: If true, flag is set.
     * - Shows message when session is timed out
     * - Forwards user to to localized shop.
     */
    function handleSubsequentAppStartups() {
        // Read AutoLoginTimeout from App Landing Page
        var loginTimeoutFromServer = $('#app-auto-login-timeout').attr('data-app-auto-login-timeout');
        var autoLoginTimeout = !isNaN(loginTimeoutFromServer) ? loginTimeoutFromServer : 30;

        getAppSettings(function (err, settings) {
            if (!err) {
                settings.autoLoginTimeout = parseInt(autoLoginTimeout);
                if (settings.lastStartup !== 0 && (new Date().getTime() - settings.lastStartup > (settings.autoLoginTimeout * 24 * 60 * 60 * 1000 ))) {
                    app.log.info('App was not used for more than ' + settings.autoLoginTimeout + ' days. Autologin is disabled.');
                    app.autoLogin.setExpired(true);
                    settings.countForSessionTimeoutMessage = 1;
                    if (settings.fingerprintLogin) {
                        app.autoLogin.setShowFingerprintWhenExpired(true);
                    }
                } else {
                    app.autoLogin.setExpired(false);
                    settings.countForSessionTimeoutMessage = 0;
                }
                // #### Popup Session destroyed ###
                if (settings.countForSessionTimeoutMessage === 1 && settings.credentials.username && settings.credentials.password && settings.credentials.valid) {
                    app.dialogs.showNotificationAlert('Session abgelaufen', 'Bitte melden Sie sich erneut an.', 'OK');
                    settings.countForSessionTimeoutMessage = 0;
                }
                settings.lastStartup = new Date().getTime();

                updateAppSettings(settings, function (err, updatedObj) {
                    if (err) {
                        app.log.error('handleSubsequentAppStartups: Unable to update appSettings. Error -> ' + JSON.stringify(err));
                    }
                    handleCookieForLocalizedShop(updatedObj);
                    forwardToLocalizedShop(updatedObj);
                });
            } else {
                app.log.error('handleSubsequentAppStartups: Unable to read appSettings. error -> ' + JSON.stringify(err));
            }
        });
    }

    /**
     * Sets a cookie containing the selected shop for analytics purposes
     * @param settings
     */
    function handleCookieForLocalizedShop(settings) {
        var cookieName = 'selected_shop';
        var cookieExpiration = 1000; // days
        if (readCookie(cookieName) === null) {
            createCookie(cookieName, settings.selectedShop.toLowerCase(), cookieExpiration);
        } else {
            eraseCookie(cookieName);
            createCookie(cookieName, settings.selectedShop.toLowerCase(), cookieExpiration);
        }
    }

    /**
     * Forwards user to country/shop specified in app settings
     * @param settings
     */
    function forwardToLocalizedShop(settings) {
        var locale = settings.selectedShop;
        if (locale) {
            window.location.replace(getBaseUrlForLocale(locale) + shopUrlHome);
        } else {
            initUniversalLinks(settings);
            window.location.replace(getBaseUrlForLocale('de') + shopUrlHome);
        }
    }

    /**
     * Respects only the region from a locale string (i.e. function will return AT for en-AT)
     * @param locale
     * @returns {*}
     */
    function getSelectedShopForLocale(locale) {
        var region = locale.split('-');
        if (region.length === 2) {
            if (region[1].toLowerCase() === 'at') {
                return 'at';
            }
        }
        return 'de';
    }

    /**
     * Reads the locale string from device settings (e.g. de-DE, en-US, de-AT, ... )
     * @param cb
     */
    function getLocaleFromDevice(cb) {
        navigator.globalization.getLocaleName(function(locale) {
                cb(null, locale.value);
            },
            function(err){
                cb(err, null);
            });
    }

    /**
     * Checks if appSettings can be loaded from device's native storage. Returns "true" if appSettings were found, otherwise "false"
     * @param cb
     */
    function checkAppSettings(cb) {
        getAppSettings(function(error, settings) {
            if (!error) {
                cb(null, true);
            } else {
                if (error.code === 2) { // error.code = 2 -> requested object was not found
                    cb(null, false);
                } else {
                    cb(error, false);
                }
            }
        });
    }

    /**
     * @param cb
     */
    function createAppSettings(cb) {
        updateAppSettings(appSettingsEmpty, cb);
    }

    /**
     * @param cb
     */
    function getAppSettings(cb) {
        window.NativeStorage.getItem(keyAppSettings, function (data) {
                cb(null, data);
            },
            function (error) {
                cb(error, null);
            });
    }

    /**
     *
     * @param settingsObj
     * @param cb
     */
    function updateAppSettings(settingsObj, cb) {
        window.NativeStorage.setItem(keyAppSettings, settingsObj,
            function (persistedObj) {
                cb(null, persistedObj);
            },
            function (error) {
                cb(error, null);
            });
    }

    // function getBaseUrlFromManifest() {
    //      window.hostedwebapp.getManifest(function(manifest) {
    //         alert('Loaded Manifest -> ' + JSON.stringify(JSON.parse(manifest).start_url));
    //     },
    //      function(error){
    //         alert('error loading manifest -> ' + JSON.stringify(error));
    //     });
    // }

    /**
     *
     * @param url
     * @returns {boolean}
     */
    function isUrlWithInvalidLocale(url) {
        var selectedShop = readCookie('selected_shop');
        var regex = /webshop30_(\S\S)/g;

        var match = regex.exec(url);
        app.log.info('isUrlWithInvalidLocale(): url: '+url);
        app.log.info('isUrlWithInvalidLocale(): match: '+match);
        if(match){
            var localeFromUrl = match[1];

            if(localeFromUrl.toLowerCase() !== selectedShop.toLowerCase()){
                app.log.debug("isUrlWithInvalidLocale() match / return true");
                return true;
            } else {
                app.log.debug("isUrlWithInvalidLocale() match / return false");
                return false;
            }
        }
        // If (austrian) user gets redirected into root directory (URL without country path, root shows german shop),
        // user gets redirected into austrian AT shop.
        if(window.location.pathname === "/"){
            if(!(selectedShop.toLowerCase()==="de")){
                app.log.debug("isUrlWithInvalidLocale() pathname AT / return true");
                return true;
            }
        }
    }

    /**
     *
     * @returns {string}
     */
    function getBaseUrl() {
        return getBaseUrlForLocale(readCookie('selected_shop'));
    }

    /**
     * @param locale
     * @returns {string}
     */
    function getBaseUrlForLocale(locale) {
        var hostname = window.location.hostname;
        var proto = window.location.protocol;
        var baseUrl = proto + '//' + hostname + '/content/webshop30_';

        if (locale && locale.toLowerCase() === 'at'){
            baseUrl = baseUrl + 'at/';
        } else {
            baseUrl = baseUrl + 'de/';
        }
        return baseUrl;
    }

    /**
     *
     * @param name
     * @param value
     * @param days
     */
    function createCookie(name, value, days) {
        var expires;
        var date;
        if (days) {
            date = new Date();
            date.setTime(date.getTime() + (days * 24 * 60 * 60 * 1000));
            expires = '; expires=' + date.toUTCString();
        }
        else {
            expires = '';
        }
        document.cookie = name + '=' + value + expires + '; path=/';
    }

    /**
     *
     * @param name
     * @returns {*}
     */
    function readCookie(name) {
        var nameEQ = name + '=';
        var ca = document.cookie.split(';');
        for(var i = 0; i < ca.length; i++) {
            var c = ca[i];
            while (c.charAt(0) === ' ') {
                c = c.substring(1, c.length);
            }
            if (c.indexOf(nameEQ) === 0) {
                return c.substring(nameEQ.length,c.length);
            }
        }
        return null;
    }

    /**
     *
     * @param name
     */
    function eraseCookie(name) {
        createCookie(name, '', -1);
    }
    /**
     * Checks if current page is app landingpage by url
     * @param {string} currentUrl
     * @returns {boolean}
     */
    function isAppLandingPage(currentUrl){
        return (currentUrl.indexOf('/app.html') > -1); // If > -1: Page is app.html
    }
    /**
     * Checks if current page is app settings page by body class
     * @returns {boolean}
     */
    function isAppSettingsPage(){
        if ($("body").hasClass(bodyClassAppSettings)){
            return true;
        } else {
            return false;
        }
    }
    /**
     * Checks if current page is check out page
     * @returns {boolean}
     */
    function isPageCheckoutAccount(){
        var currentUrl = window.location.href;
        return (currentUrl.indexOf('/checkout.account.html') > -1); // If > -1: Page is checkout.account.html
    }

    /**
     *
     * @returns {boolean}
     */
    function isShopHomepage(){
        var currentUrl = window.location.href;
        return (currentUrl.indexOf('/' + shopUrlHome) > -1); // If > -1: Page is app.html
    }

    /**
     * @returns {string}
     */
    function getCurrentUrl(){
        return window.location.href;
    }

    /**
     * Function shows ask for fingerprint dialog and saves decision in app settings
     * @param {} userSettings Local app settings
     */
    function handleAskForFingerprintDecisionDialog(userSettings){
        isFingerprintScannerAvailable(function (err, res) {
            if (!err) {
                if (res.isAvailable) {
                    var settings = userSettings;
                    if ((settings.fingerprintLoginDecision === '' || (settings.fingerprintLoginDecision === 'later' && settings.showDialogAskForFingerprintDecision)) && !settings.fingerprintLogin && isUserLoggedIn()) {
                        settings.showDialogAskForFingerprintDecision = false; //disable dialog for this session
                        app.fingerprintLogin.askForFingerprint(function (buttonIdx) {
                            switch (buttonIdx) {
                                case 1: { // Yes
                                    window.ADB.trackAction("aktivierung.Fingerprint",{"status.aktivierung.fingerprint": "aktivieren"});
                                    settings.fingerprintLogin = true;
                                    settings.fingerprintLoginDecision = 'yes';
                                    settings.showDialogAskForFingerprintDecision = false;
                                    break;
                                }
                                case 2: { // Later
                                    window.ADB.trackAction("aktivierung.Fingerprint",{"status.aktivierung.fingerprint": "spaeter"});
                                    //settings.fingerprintLogin = true;
                                    settings.fingerprintLoginDecision = 'later';
                                    settings.showDialogAskForFingerprintDecision = false;
                                    break;
                                }
                                case 3: { // No
                                    window.ADB.trackAction("aktivierung.Fingerprint",{"status.aktivierung.fingerprint": "nein"});
                                    settings.fingerprintLogin = false;
                                    settings.fingerprintLoginDecision = 'no';
                                    settings.showDialogAskForFingerprintDecision = false;
                                    break;
                                }
                            }
                            updateAppSettings(settings, function (err, res) {
                                if (err) {
                                    app.log.error('handleAskForFingerprintLogin(): 1 Error persisting appSettings. err -> ' + JSON.stringify(err));
                                }
                            });
                        });
                    }
                }
            }
        });
    }

    /**
     * Used to SET whether fingerprintlogin attempt was performed to enable/disable further fingerprintlogin dialogs
     * @param bool
     */
    function setFlagFingerprintloginAttemptPerformed(bool) {
        window.sessionStorage.setItem('fingerprintLoginAttemptPerformed1', '' + bool);
    }

    /**
     * Used to GET whether fingerprintlogin attempt was performed or not
     * @returns {boolean}
     */
    function isFlagFingerprintLoginAttemptPerformed() {
        var fplPerformed = window.sessionStorage.getItem('fingerprintLoginAttemptPerformed1');
        return (fplPerformed === 'true') ?  true : false;
    }
    /**
     * Function for filling and handling page "app settings". Handles:
     * - Switch Shop/Country
     * - Switch Autologin
     * - Switch Fingerprint login
     * - App Rating Button
     * - Sets App Version
     */
    var appSettingsPageUi = {
        /**
         * Fills country switch on app settings
         * @param settings
         */
        setUiShopSwitchCountry: function(settings){
            if (settings.selectedShop === 'de') {
                $('#selectShop').val('de');
                $('#selectShopSelectBoxItText').text('Deutschland');
            } else if (settings.selectedShop === 'at') {
                $('#selectShop').val('at');
                $('#selectShopSelectBoxItText').text('Österreich');
            } else {
                app.log.error('Unexpected value in app settings for property "selectedShop: ' + settings.selectedShop);
            }
        },
        /**
         * Handles changes on country switch on app settings
         * @param settings
         */
        handleSwitchShopCountry: function(settings){
            $('#selectShop').change(function () {
                settings.selectedShop = $(this).val();
                updateAppSettings(settings, function (err, res) {
                    if (err) {
                        app.log.error('initAppSettingsPage unable to persist settings. err -> ' + JSON.stringify(err));
                    } else {
                        window.ADB.trackAction("setting.Countryselection", {"country":settings.selectedShop});
                        handleCookieForLocalizedShop(settings);
                        window.location.replace(getBaseUrlForLocale(res.selectedShop) + shopUrlAppSettings);
                    }
                });
            });
        },
        /**
         * Sets autologin switch on app settings page
         * @param settings
         */
        setUiAutologin: function(settings){
            if (settings.autoLogin) {
                $('#switchAutologing').attr('checked', true);
            } else if (!settings.autoLogin) {
                $('#switchAutologing').attr('checked', false);
            } else {
                app.log.error('Unexpected value in app settings for property "autoLogin: ' + settings.autoLogin);
            }
        },
        /**
         * Handles changes on autologin switch on app settings page
         * @param settings
         */
        handleSwitchAutologin: function(settings){
            $('#switchAutologing').change(function () {
                settings.autoLogin = !settings.autoLogin;
                updateAppSettings(settings, function (err, res) {
                    if (err) {
                        app.log.error('initAppSettingsPage unable to persist autoLogin settings. err -> ' + JSON.stringify(err));
                    }
                    if(res){
                        if(settings.autoLogin){
                            window.ADB.trackAction("setting.Autologin", {"status.Autologin":"aktivieren"});
                            app.log.debug("adb settings.autoLogin: "+settings.autoLogin);
                        } else {
                            window.ADB.trackAction("setting.Autologin", {"status.Autologin":"deaktivieren"});
                            app.log.debug("adb settings.autoLogin: "+settings.autoLogin);
                        }
                    }
                });
            });
        },
        /**
         * Sets Fingerprintlogin Switch on app settings page
         * @param settings
         * @param fingerPrintScanner
         */
        setUiFingerprintLogin: function (settings, fingerPrintScanner) {
            if (!fingerPrintScanner.isHardwareDetected) {
                $('#fingerprintBtn').parent().next().css('display', 'none');
                $('#fingerprintBtn').parent().css('display', 'none');
            } else {
                if(isIos()){
                    if(fingerPrintScanner.type==="face"){
                        $('.icon-fingerprint').css('display', 'none');
                        $('.menuBtnText.ios-faceid').css('display', 'block');
                        if(!fingerPrintScanner.isAvailable){
                            $('.appMenuHelperText.copyBig.ios-faceid.isConfigured').css('display', 'block');
                        } else {
                            $('.appMenuHelperText.copyBig.ios-faceid:not(.isConfigured)').css('display', 'block');
                        }
                    }
                    else if(isIphoneX()){
                        $('.icon-fingerprint').css('display', 'none');
                        $('.menuBtnText.ios-faceid').css('display', 'block');
                        $('.appMenuHelperText.copyBig.ios-faceid.isConfigured').css('display', 'block');
                    }
                    else {
                        $('.menuBtnText.ios-fingerprint').css('display', 'block');
                        if(!fingerPrintScanner.isAvailable){
                            $('.appMenuHelperText.copyBig.ios-fingerprint.isConfigured').css('display', 'block');
                        } else {
                            $('.appMenuHelperText.copyBig.ios-fingerprint:not(.isConfigured)').css('display', 'block');
                        }
                    }
                } else if(isAndroid()){
                    $('.menuBtnText.android-fingerprint').css('display', 'block');
                    if(!fingerPrintScanner.isAvailable){
                        $('.appMenuHelperText.copyBig.android-fingerprint.isConfigured').css('display', 'block');
                    } else {
                        $('.appMenuHelperText.copyBig.android-fingerprint:not(.isConfigured)').css('display', 'block');
                    }
                }
                if(settings.fingerprintLogin){
                    $('#switchFingerprint').attr('checked', true);
                } else {
                    $('#switchFingerprint').attr('checked', false);
                }
                if(!fingerPrintScanner.isAvailable){
                    $('#fingerprintBtn').parent().next().find(".appMenuHelperText").text('Der Fingerprint Sensor muss erst in den Smartphone-Einstellungen konfiguriert werden, bevor dieses Feature genutzt werden kann!');
                }
            }
        },
        /**
         * Handles changes on fingerprintlogin switch on app settings page
         * @param settings
         * @param fingerPrintScanner
         */
        handleSwitchFingerprintLogin: function(settings, fingerPrintScanner){
            if (fingerPrintScanner.isHardwareDetected) {
                $('#switchFingerprint').change(function(){
                    if($('#switchFingerprint').is(':checked')){
                        settings.fingerprintLogin = true;
                    } else {
                        settings.fingerprintLogin = false;
                    }
                    settings.fingerprintLoginDecision = 'yes';
                    updateAppSettings(settings, function (err, res) {
                        if(err) {
                            app.log.error('initAppSettingsPage/handleSwitchFingerprintLogin unable to persist fingerprintSettings. err -> ' + JSON.stringify(err));
                        } else {
                            if(res.fingerprintLogin){
                                window.ADB.trackAction("setting.Fingerprint", {"status.Fingerprint":"aktivieren"});
                                app.log.debug("adb res.fingerprintLogin: "+res.fingerprintLogin);
                            } else {
                                window.ADB.trackAction("setting.Fingerprint", {"status.Fingerprint":"deaktivieren"});
                                app.log.debug("adb res.fingerprintLogin: "+res.fingerprintLogin);
                            }

                            if(res.fingerprintLogin && !isUserLoggedIn()){
                                getAppSettings(function (err, settings) {
                                    if (!err) {
                                        if(settings.fingerprintLogin && settings.credentials.username && settings.credentials.password && settings.credentials.valid){
                                            verifyFingerprint(function(err, fingerprintIsValid) {
                                                app.log.debug('attempt to logging in after touchID verification');
                                                if (!err) {
                                                    if (fingerprintIsValid === true) {
                                                        var loginComponent = getLoginComponentHeaderBar();

                                                        // add handler to remove loading animation of login component
                                                        app.loginFailedHandler.add('errorDefaultLogin', function(){
                                                            window.animationHelper.removeLoadingAnimation(loginComponent.find('.component.login'));
                                                        });

                                                        loginWithInvisibleForm(loginComponent, settings.credentials.username, settings.credentials.password, "fingerprint");
                                                    }
                                                } else {
                                                    app.log.error('AppSettings FingerprintLogin: Error trying to verify fingerprint. err -> ' + JSON.stringify(err));
                                                }
                                            });
                                        }
                                    } else {
                                        app.log.error('AppSettings FingerprintLogin: Unable to read appSettings. error -> ' + JSON.stringify(err));
                                    }
                                });
                            }
                        }
                    });
                });
            }
        },
        /**
         * Inserts app version on app settings page
         */
        setUiAppVersion: function(){
            var tmpAppVersion;
            if (isAndroid()) {
                tmpAppVersion = appVersionAndroid + ' (Android)';
            } else if (isIos()) {
                tmpAppVersion = appVersionIos + ' (iOS)';
            } else {
                tmpAppVersion = '(' + appVersionAndroid + ' / ' + appVersionIos + ')';
            }
            $('.appSettingsMenu').after('<p style="padding-left:12px;">App-Version: ' + tmpAppVersion + '</p>');
        },
        /**
         * Binds function on app rating button on app settings page
         */
        initAppRatingPopUpFunction: function (){
            // Bind app rating click event
            var appRatingButton = $('#rateAppBtn');
            if (appRatingButton.length === 1) {
                $(appRatingButton).on('click', function(event){
                    app.triggerAppRatingOpenMenuEvent();
                });
            }
        }
    }
    app.initialize();
})($);